@@ -12,9 +12,8 @@ from ambulance_game.simulation.simulation import (
     build_custom_node,
     build_model,
     simulate_model,
-    calculate_class_2_individuals_best_response,
     get_average_simulated_state_probabilities,
-    get_mean_blocking_difference_between_two_systems,
+    get_mean_blocking_difference_using_simulation,
     get_multiple_runs_results,
     get_simulated_state_probabilities,
     extract_total_individuals_and_the_ones_within_target_for_both_classes,
@@ -39,6 +38,30 @@ def test_build_model(lambda_2, lambda_1, mu, c):
     assert isinstance(result, ciw.network.Network)
 
 
+def test_build_model_fair_allocation():
+    """
+    Test that the correct busy times for servers are output given specific
+    values and fair allocation of individuals
+    """
+
+    def server_busy_time_priority(srv, ind):  # pylint: disable=unused-argument
+        return srv.busy_time
+
+    network = build_model(
+        lambda_2=1,
+        lambda_1=1,
+        mu=2,
+        num_of_servers=2,
+        server_priority_function=server_busy_time_priority,
+    )
+    ciw.seed(0)
+    Q = ciw.Simulation(network)
+    Q.simulate_until_max_time(100)
+    expected_busy_times = [47.813604079373505, 48.03271051578811]
+    simulated_busy_times = [srv.busy_time for srv in Q.nodes[2].servers]
+    assert np.allclose(simulated_busy_times, expected_busy_times)
+
+
 def test_example_model():
     """
     Test to ensure that the correct results are output to a specific problem
@@ -135,9 +158,9 @@ def test_simulate_model_unconstrained():
         )
         rec = simulation.get_all_records()
         sim_results.append(rec)
-        blocks = blocks + sum([b.time_blocked for b in rec])
-        waits = waits + sum([w.waiting_time for w in rec])
-        services = services + sum([s.service_time for s in rec])
+        blocks = blocks + sum(b.time_blocked for b in rec)
+        waits = waits + sum(w.waiting_time for w in rec)
+        services = services + sum(s.service_time for s in rec)
 
     assert isinstance(simulation, ciw.simulation.Simulation)
     assert len(sim_results[0]) == 474
@@ -178,9 +201,9 @@ def test_simulate_model_constrained():
         )
         rec = simulation.get_all_records()
         sim_results.append(rec)
-        blocks = blocks + sum([b.time_blocked for b in rec])
-        waits = waits + sum([w.waiting_time for w in rec])
-        services = services + sum([s.service_time for s in rec])
+        blocks = blocks + sum(b.time_blocked for b in rec)
+        waits = waits + sum(w.waiting_time for w in rec)
+        services = services + sum(s.service_time for s in rec)
 
     assert isinstance(simulation, ciw.simulation.Simulation)
     assert len(sim_results[0]) == 504
@@ -587,9 +610,7 @@ def test_example_get_multiple_results_for_different_individuals_classes():
     lambda_1=floats(min_value=0.1, max_value=0.4),
 )
 @settings(deadline=None)
-def test_get_mean_blocking_difference_between_two_systems_equal_split(
-    lambda_2, lambda_1
-):
+def test_get_mean_blocking_difference_using_simulation_equal_split(lambda_2, lambda_1):
     """
     Test that ensures that the function that finds the optimal distribution of
     class 2 individuals in two identical systems returns a solution that
@@ -598,13 +619,13 @@ def test_get_mean_blocking_difference_between_two_systems_equal_split(
     is precisely what the function checks. This test runs the function with a
     proportion variable of 0.5 (meaning equally distributing class 2 individuals
     between the two systems) and ensures that the difference is 0, given any
-    values of λ^α and λ^ο_1 = λ^ο_2 = λ^ο
+    values of lambda_2 and lambda_1_1 = lambda_1_2 = lambda_1.
 
     Note here that due to the ciw.seed() function it was possible to eliminate any
     randomness and make both systems identical, in terms of arrivals, services
     and any other stochasticity that the simulation models incorporates.
     """
-    diff = get_mean_blocking_difference_between_two_systems(
+    diff = get_mean_blocking_difference_using_simulation(
         prop_1=0.5,
         lambda_2=lambda_2,
         lambda_1_1=lambda_1,
@@ -629,13 +650,13 @@ def test_get_mean_blocking_difference_between_two_systems_equal_split(
 
 
 # TODO Investigate making this a property based test
-def test_get_mean_blocking_difference_between_two_systems_increasing():
+def test_get_mean_blocking_difference_using_simulation_increasing():
     """Ensuring that the function is increasing for specific inputs"""
     diff_list = []
     proportions = np.linspace(0.1, 0.9, 9)
     for prop in proportions:
         diff_list.append(
-            get_mean_blocking_difference_between_two_systems(
+            get_mean_blocking_difference_using_simulation(
                 prop_1=prop,
                 lambda_2=0.15,
                 lambda_1_1=0.08,
@@ -661,86 +682,37 @@ def test_get_mean_blocking_difference_between_two_systems_increasing():
     assert is_increasing
 
 
-#  TODO Investigate making it a property based test
-def test_calculate_class_2_individuals_best_response_equal_split():
-    """Make sure that the brentq() function that is used suggests that when two
-    identical systems are considered the individuals will be split equally between
-    them (50% - 50%)
-
-    Note here that due to the ciw.seed() function it was possible to eliminate any
-    randomness and make both systems identical, in terms of arrivals, services
-    and any other stochasticity that the simulation models incorporates.
-    """
-    lambda_2 = 0.3
-    equal_split = calculate_class_2_individuals_best_response(
-        lambda_2=lambda_2,
-        lambda_1_1=0.3,
-        lambda_1_2=0.3,
-        mu_1=0.2,
-        mu_2=0.2,
-        num_of_servers_1=4,
-        num_of_servers_2=4,
-        threshold_1=3,
-        threshold_2=3,
-        seed_num_1=10,
-        seed_num_2=10,
-        num_of_trials=5,
-        warm_up_time=100,
-        runtime=500,
-        system_capacity_1=float("inf"),
-        system_capacity_2=float("inf"),
-        buffer_capacity_1=float("inf"),
-        buffer_capacity_2=float("inf"),
-    )
-
-    assert np.isclose(equal_split, 0.5)
-
-
-def test_calculate_class_2_individuals_best_response_all_individuals_in_one():
-    """
-    Ensuring that the function is sends 100% of individuals to the first system
-    when the second system is very busy and vise versa.
-    """
-    all_individuals_to_first = calculate_class_2_individuals_best_response(
-        lambda_2=0.3,
-        lambda_1_1=0.1,
-        lambda_1_2=3,
-        mu_1=10,
-        mu_2=2,
-        num_of_servers_1=8,
-        num_of_servers_2=4,
-        threshold_1=6,
-        threshold_2=3,
-        seed_num_1=10,
-        seed_num_2=10,
-        num_of_trials=5,
-        warm_up_time=100,
-        runtime=500,
-        system_capacity_1=float("inf"),
-        system_capacity_2=float("inf"),
-        buffer_capacity_1=float("inf"),
-        buffer_capacity_2=float("inf"),
-    )
-    assert all_individuals_to_first == 1
-
-    all_individuals_to_second = calculate_class_2_individuals_best_response(
-        lambda_2=0.3,
-        lambda_1_1=3,
-        lambda_1_2=0.1,
-        mu_1=2,
-        mu_2=10,
-        num_of_servers_1=4,
-        num_of_servers_2=8,
-        threshold_1=3,
-        threshold_2=6,
-        seed_num_1=10,
-        seed_num_2=10,
-        num_of_trials=5,
-        warm_up_time=100,
-        runtime=500,
-        system_capacity_1=float("inf"),
-        system_capacity_2=float("inf"),
-        buffer_capacity_1=float("inf"),
-        buffer_capacity_2=float("inf"),
-    )
-    assert all_individuals_to_second == 0
+def test_get_mean_blocking_difference_using_bounded_simulation_increasing():
+    """Ensuring that the function is increasing for specific inputs"""
+    expected_out = [
+        -5.920904907479688,
+        -2.1671852642543246,
+        2.7733256111964444,
+        7.009536618576061,
+    ]
+    proportions = np.linspace(0.2, 0.8, 4)
+    for index, prop in enumerate(proportions):
+        mean_diff = get_mean_blocking_difference_using_simulation(
+            prop_1=prop,
+            lambda_2=0.15,
+            lambda_1_1=0.08,
+            lambda_1_2=0.08,
+            mu_1=0.05,
+            mu_2=0.05,
+            num_of_servers_1=6,
+            num_of_servers_2=6,
+            threshold_1=5,
+            threshold_2=5,
+            seed_num_1=2,
+            seed_num_2=2,
+            num_of_trials=100,
+            warm_up_time=100,
+            runtime=500,
+            system_capacity_1=10,
+            system_capacity_2=15,
+            buffer_capacity_1=8,
+            buffer_capacity_2=6,
+        )
+        assert round(mean_diff, NUMBER_OF_DIGITS_TO_ROUND) == round(
+            expected_out[index], NUMBER_OF_DIGITS_TO_ROUND
+        )
